@@ -4,7 +4,7 @@
  * Copyright (C) 2001 Havoc Pennington
  * Copyright (C) 2002 Red Hat, Inc.
  * Copyright (C) 2003 Rob Adams
- * Copyright (C) 2004, 2005 Elijah Newren
+ * Copyright (C) 2004-2006 Elijah Newren
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -33,6 +33,7 @@
 #include <X11/Xlib.h>
 #include "eventqueue.h"
 #include "common.h"
+#include "boxes.h"
 
 #ifdef HAVE_STARTUP_NOTIFICATION
 #include <libsn/sn.h>
@@ -44,17 +45,6 @@
 
 #define meta_XFree(p) do { if ((p)) XFree ((p)); } while (0)
 
-/* this doesn't really belong here, oh well. */
-typedef struct _MetaRectangle MetaRectangle;
-
-struct _MetaRectangle
-{
-  int x;
-  int y;
-  int width;
-  int height;
-};
-
 typedef struct MetaCompositor  MetaCompositor;
 typedef struct _MetaDisplay    MetaDisplay;
 typedef struct _MetaFrame      MetaFrame;
@@ -68,6 +58,8 @@ typedef struct _MetaWorkspace  MetaWorkspace;
 typedef struct _MetaWindowPropHooks MetaWindowPropHooks;
 typedef struct _MetaGroupPropHooks  MetaGroupPropHooks;
 
+typedef struct MetaEdgeResistanceData MetaEdgeResistanceData;
+
 typedef void (* MetaWindowPingFunc) (MetaDisplay *display,
 				     Window       xwindow,
 				     Time         timestamp,
@@ -212,6 +204,14 @@ struct _MetaDisplay
    */
   guint mouse_mode : 1;
 
+  /* Helper var for strict focus for terminals; only relevant if the focus
+   * window is a terminal.  Typically, we don't allow new windows to take
+   * focus away from a terminal, but if the user explicitly did something
+   * that should allow a different window to gain focus (e.g. global
+   * keybinding or clicking on a dock), then we will allow the transfer.
+   */
+  guint allow_terminal_deactivation : 1;
+
   guint static_gravity_works : 1;
   
   /*< private-ish >*/
@@ -224,11 +224,6 @@ struct _MetaDisplay
                               XErrorEvent *error);  
   int server_grab_count;
 
-  /* This window holds the focus when we don't want to focus
-   * any actual clients
-   */
-  Window no_focus_window;
-  
   /* for double click */
   Time last_button_time;
   Window last_button_xwindow;
@@ -274,12 +269,15 @@ struct _MetaDisplay
   MetaRectangle grab_wireframe_rect;
   MetaRectangle grab_wireframe_last_xor_rect;
   MetaRectangle grab_initial_window_pos;
+  int         grab_initial_x, grab_initial_y;  /* These are only relevant for */
+  gboolean    grab_threshold_movement_reached; /* raise_on_click == FALSE.    */
   MetaResizePopup *grab_resize_popup;
   GTimeVal    grab_last_moveresize_time;
   Time        grab_motion_notify_time;
   int         grab_wireframe_last_display_width;
   int         grab_wireframe_last_display_height;
   GList*      grab_old_window_stacking;
+  MetaEdgeResistanceData *grab_edge_resistance_data;
   
   /* we use property updates as sentinels for certain window focus events
    * to avoid some race conditions on EnterNotify events
@@ -288,6 +286,7 @@ struct _MetaDisplay
 
 #ifdef HAVE_XKB
   int         xkb_base_event_type;
+  Time        last_bell_time;
 #endif
 #ifdef HAVE_XSYNC
   /* alarm monitoring client's _NET_WM_SYNC_REQUEST_COUNTER */
@@ -350,6 +349,7 @@ struct _MetaDisplay
   int render_error_base;
 #endif
 #ifdef HAVE_XSYNC
+  unsigned int grab_last_user_action_was_snap;
   unsigned int have_xsync : 1;
 #define META_DISPLAY_HAS_XSYNC(display) ((display)->have_xsync)
 #else
@@ -382,13 +382,13 @@ struct _MetaDisplay
  * the result.
  */
 #define XSERVER_TIME_IS_BEFORE_ASSUMING_REAL_TIMESTAMPS(time1, time2) \
-  ( (( time1 < time2 ) && ( time2 - time1 < ((guint32)-1)/2 )) ||     \
-    (( time1 > time2 ) && ( time1 - time2 > ((guint32)-1)/2 ))        \
+  ( (( (time1) < (time2) ) && ( (time2) - (time1) < ((guint32)-1)/2 )) ||     \
+    (( (time1) > (time2) ) && ( (time1) - (time2) > ((guint32)-1)/2 ))        \
   )
 #define XSERVER_TIME_IS_BEFORE(time1, time2)                          \
-  ( time1 == 0 ||                                                     \
+  ( (time1) == 0 ||                                                     \
     (XSERVER_TIME_IS_BEFORE_ASSUMING_REAL_TIMESTAMPS(time1, time2) && \
-     time2 != 0)                                                      \
+     (time2) != 0)                                                      \
   )
 
 gboolean      meta_display_open                (const char  *name);
@@ -419,6 +419,9 @@ void        meta_display_register_x_window   (MetaDisplay *display,
                                               MetaWindow  *window);
 void        meta_display_unregister_x_window (MetaDisplay *display,
                                               Window       xwindow);
+/* Return whether the xwindow is a no focus window for any of the screens */
+gboolean    meta_display_xwindow_is_a_no_focus_window (MetaDisplay *display,
+                                                       Window xwindow);
 
 GSList*     meta_display_list_windows        (MetaDisplay *display);
 
@@ -449,6 +452,9 @@ gboolean meta_display_begin_grab_op (MetaDisplay *display,
 void     meta_display_end_grab_op   (MetaDisplay *display,
                                      Time         timestamp);
 
+void    meta_display_check_threshold_reached (MetaDisplay *display,
+                                              int          x,
+                                              int          y);
 void     meta_display_grab_window_buttons    (MetaDisplay *display,
                                               Window       xwindow);
 void     meta_display_ungrab_window_buttons  (MetaDisplay *display,
@@ -459,6 +465,10 @@ void meta_display_grab_focus_window_button   (MetaDisplay *display,
 void meta_display_ungrab_focus_window_button (MetaDisplay *display,
                                               MetaWindow  *window);
 
+/* Next two functions are defined in edge-resistance.c */
+void meta_display_compute_resistance_and_snapping_edges (MetaDisplay *display);
+void meta_display_cleanup_edges                         (MetaDisplay *display);
+
 /* make a request to ensure the event serial has changed */
 void     meta_display_increment_event_serial (MetaDisplay *display);
 
@@ -490,10 +500,14 @@ typedef enum
 {
   META_TAB_LIST_NORMAL,
   META_TAB_LIST_DOCKS
-  
-
 } MetaTabList;
 
+typedef enum
+{
+  META_TAB_SHOW_ICON,      /* Alt-Tab mode */
+  META_TAB_SHOW_INSTANTLY  /* Alt-Esc mode */
+} MetaTabShowType;
+
 GList* meta_display_get_tab_list (MetaDisplay   *display,
                                   MetaTabList    type,
                                   MetaScreen    *screen,
@@ -516,12 +530,6 @@ int meta_resize_gravity_from_grab_op (MetaGrabOp op);
 gboolean meta_grab_op_is_moving   (MetaGrabOp op);
 gboolean meta_grab_op_is_resizing (MetaGrabOp op);
 
-gboolean meta_rectangle_intersect (MetaRectangle *src1,
-                                   MetaRectangle *src2,
-                                   MetaRectangle *dest);
-gboolean meta_rectangle_equal (const MetaRectangle *src1,
-                               const MetaRectangle *src2);
-
 void meta_display_devirtualize_modifiers (MetaDisplay        *display,
                                           MetaVirtualModifier modifiers,
                                           unsigned int       *mask);
@@ -549,6 +557,7 @@ void meta_display_set_input_focus_window   (MetaDisplay *display,
  * same as meta_display_set_input_focus_window
  */
 void meta_display_focus_the_no_focus_window (MetaDisplay *display, 
+                                             MetaScreen  *screen,
                                              Time         timestamp);
 
 void meta_display_queue_autoraise_callback  (MetaDisplay *display,
