@@ -2,6 +2,8 @@
 
 /* 
  * Copyright (C) 2001 Havoc Pennington
+ * Copyright (C) 2003 Red Hat, Inc.
+ * Copyright (C) 2005, 2006 Elijah Newren
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -61,7 +63,12 @@ static gboolean meta_frames_leave_notify_event    (GtkWidget           *widget,
 static void meta_frames_paint_to_drawable (MetaFrames   *frames,
                                            MetaUIFrame  *frame,
                                            GdkDrawable  *drawable,
-                                           GdkRegion    *region);
+                                           GdkRegion    *region,
+					   int           x_offset,
+					   int           y_offset);
+
+static void meta_frames_set_window_background (MetaFrames   *frames,
+                                               MetaUIFrame  *frame);
 
 static void meta_frames_calc_geometry (MetaFrames        *frames,
                                        MetaUIFrame         *frame,
@@ -84,6 +91,9 @@ static MetaFrameControl get_control  (MetaFrames        *frames,
                                       int                x,
                                       int                y);
 static void clear_tip (MetaFrames *frames);
+static void invalidate_all_caches (MetaFrames *frames);
+static void invalidate_whole_window (MetaFrames *frames,
+				     MetaUIFrame *frame);
 
 static GtkWidgetClass *parent_class = NULL;
 
@@ -155,7 +165,7 @@ unsigned_long_hash (gconstpointer v)
   gulong val = * (const gulong *) v;
 
   /* I'm not sure this works so well. */
-#if G_SIZEOF_LONG > 4
+#if GLIB_SIZEOF_LONG > 4
   return (guint) (val ^ (val >> 32));
 #else
   return val;
@@ -184,7 +194,7 @@ meta_frames_init (MetaFrames *frames)
 {
   GTK_WINDOW (frames)->type = GTK_WINDOW_POPUP;
 
-  frames->text_heights = g_hash_table_new (g_int_hash, g_int_equal);
+  frames->text_heights = g_hash_table_new (NULL, NULL);
   
   frames->frames = g_hash_table_new (unsigned_long_hash, unsigned_long_equal);
 
@@ -192,6 +202,10 @@ meta_frames_init (MetaFrames *frames)
 
   frames->expose_delay_count = 0;
 
+  frames->invalidate_cache_timeout_id = 0;
+  frames->invalidate_frames = NULL;
+  frames->cache = g_hash_table_new (g_direct_hash, g_direct_equal);
+
   gtk_widget_set_double_buffered (GTK_WIDGET (frames), FALSE);
 
   meta_prefs_add_listener (prefs_changed_callback, frames);
@@ -249,13 +263,97 @@ meta_frames_finalize (GObject *object)
   meta_prefs_remove_listener (prefs_changed_callback, frames);
   
   g_hash_table_destroy (frames->text_heights);
+
+  invalidate_all_caches (frames);
+  if (frames->invalidate_cache_timeout_id)
+    g_source_remove (frames->invalidate_cache_timeout_id);
   
   g_assert (g_hash_table_size (frames->frames) == 0);
   g_hash_table_destroy (frames->frames);
+  g_hash_table_destroy (frames->cache);
 
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+typedef struct
+{
+  GdkPixmap *top;
+  gint top_x, top_y;
+  
+  GdkPixmap *left;
+  gint left_x, left_y;
+  
+  GdkPixmap *right;
+  gint right_y, right_x;
+  
+  GdkPixmap *bottom;
+  gint bottom_x, bottom_y;
+  
+} CachedPixels;
+
+static CachedPixels *
+get_cache (MetaFrames *frames,
+	   MetaUIFrame *frame)
+{
+  CachedPixels *pixels;
+  
+  pixels = g_hash_table_lookup (frames->cache, frame);
+
+  if (!pixels)
+    {
+      pixels = g_new0 (CachedPixels, 1);
+      g_hash_table_insert (frames->cache, frame, pixels);
+    }
+
+  return pixels;
+}
+
+static void
+invalidate_cache (MetaFrames *frames,
+		  MetaUIFrame *frame)
+{
+  CachedPixels *pixels = get_cache (frames, frame);
+  
+  if (pixels->top)
+    g_object_unref (pixels->top);
+  if (pixels->left)
+    g_object_unref (pixels->left);
+  if (pixels->right)
+    g_object_unref (pixels->right);
+  if (pixels->bottom)
+    g_object_unref (pixels->bottom);
+  
+  g_free (pixels);
+  
+  g_hash_table_remove (frames->cache, frame);
+}
+
+static void
+invalidate_all_caches (MetaFrames *frames)
+{
+  GList *l;
+
+  for (l = frames->invalidate_frames; l; l = l->next)
+    {
+      MetaUIFrame *frame = l->data;
+
+      invalidate_cache (frames, frame);
+    }
+  
+  g_list_free (frames->invalidate_frames);
+  frames->invalidate_frames = NULL;
+}
+
+static gboolean
+invalidate_cache_timeout (gpointer data)
+{
+  MetaFrames *frames = data;
+  
+  invalidate_all_caches (frames);
+  frames->invalidate_cache_timeout_id = 0;
+  return FALSE;
+}
+
 static void
 queue_recalc_func (gpointer key, gpointer value, gpointer data)
 {
@@ -269,9 +367,9 @@ queue_recalc_func (gpointer key, gpointer value, gpointer data)
    * resize may not actually be needed so we always redraw
    * in case of color change.
    */
-  gtk_style_set_background (GTK_WIDGET (frames)->style,
-                            frame->window, GTK_STATE_NORMAL);
-  gdk_window_invalidate_rect (frame->window, NULL, FALSE);
+  meta_frames_set_window_background (frames, frame);
+  
+  invalidate_whole_window (frames, frame);
   meta_core_queue_frame_resize (gdk_display,
                                 frame->xwindow);
   if (frame->layout)
@@ -292,7 +390,7 @@ meta_frames_font_changed (MetaFrames *frames)
   if (g_hash_table_size (frames->text_heights) > 0)
     {
       g_hash_table_destroy (frames->text_heights);
-      frames->text_heights = g_hash_table_new (g_int_hash, g_int_equal);
+      frames->text_heights = g_hash_table_new (NULL, NULL);
     }
   
   /* Queue a draw/resize on all frames */
@@ -314,9 +412,9 @@ queue_draw_func (gpointer key, gpointer value, gpointer data)
    * resize may not actually be needed so we always redraw
    * in case of color change.
    */
-  gtk_style_set_background (GTK_WIDGET (frames)->style,
-                            frame->window, GTK_STATE_NORMAL);
-  gdk_window_invalidate_rect (frame->window, NULL, FALSE);
+  meta_frames_set_window_background (frames, frame);
+
+  invalidate_whole_window (frames, frame);
 }
 
 static void
@@ -391,9 +489,9 @@ meta_frames_ensure_layout (MetaFrames  *frames,
 						 meta_prefs_get_titlebar_font ());
 
       size = pango_font_description_get_size (font_desc);
-      
+
       if (g_hash_table_lookup_extended (frames->text_heights,
-                                        &size,
+                                        GINT_TO_POINTER (size),
                                         &key, &value))
         {
           frame->text_height = GPOINTER_TO_INT (value);
@@ -405,7 +503,7 @@ meta_frames_ensure_layout (MetaFrames  *frames,
                                                   gtk_widget_get_pango_context (widget));
 
           g_hash_table_replace (frames->text_heights,
-                                &size,
+                                GINT_TO_POINTER (size),
                                 GINT_TO_POINTER (frame->text_height));
         }
       
@@ -476,8 +574,9 @@ meta_frames_manage_window (MetaFrames *frames,
   frame->window = window;
 
   gdk_window_set_user_data (frame->window, frames);
-  gtk_style_set_background (GTK_WIDGET (frames)->style,
-			    frame->window, GTK_STATE_NORMAL);
+
+  /* Set the window background to the current style */
+  meta_frames_set_window_background (frames, frame);
 
   /* Don't set event mask here, it's in frame.c */
   
@@ -507,6 +606,11 @@ meta_frames_unmanage_window (MetaFrames *frames,
 
   if (frame)
     {
+      /* invalidating all caches ensures the frame
+       * is not actually referenced anymore
+       */
+      invalidate_all_caches (frames);
+      
       /* restore the cursor */
       meta_core_set_screen_cursor (gdk_display,
 				   frame->xwindow,
@@ -536,10 +640,6 @@ meta_frames_unmanage_window (MetaFrames *frames,
 static void
 meta_frames_realize (GtkWidget *widget)
 {
-  MetaFrames *frames;
-
-  frames = META_FRAMES (widget);
-  
   if (GTK_WIDGET_CLASS (parent_class)->realize)
     GTK_WIDGET_CLASS (parent_class)->realize (widget);
 }
@@ -547,10 +647,6 @@ meta_frames_realize (GtkWidget *widget)
 static void
 meta_frames_unrealize (GtkWidget *widget)
 {
-  MetaFrames *frames;
-
-  frames = META_FRAMES (widget);
-  
   if (GTK_WIDGET_CLASS (parent_class)->unrealize)
     GTK_WIDGET_CLASS (parent_class)->unrealize (widget);
 }
@@ -605,14 +701,11 @@ void
 meta_frames_reset_bg (MetaFrames *frames,
                       Window  xwindow)
 {
-  GtkWidget *widget;
   MetaUIFrame *frame;
   
-  widget = GTK_WIDGET (frames);
-
   frame = meta_frames_lookup_window (frames, xwindow);
-  
-  gtk_style_set_background (widget->style, frame->window, GTK_STATE_NORMAL);
+
+  meta_frames_set_window_background (frames, frame);
 }
 
 static void
@@ -632,11 +725,8 @@ meta_frames_unflicker_bg (MetaFrames *frames,
                           int         target_width,
                           int         target_height)
 {
-  GtkWidget *widget;
   MetaUIFrame *frame;
   
-  widget = GTK_WIDGET (frames);
-
   frame = meta_frames_lookup_window (frames, xwindow);
   g_return_if_fail (frame != NULL);
 
@@ -662,15 +752,12 @@ meta_frames_apply_shapes (MetaFrames *frames,
 {
 #ifdef HAVE_SHAPE
   /* Apply shapes as if window had new_window_width, new_window_height */
-  GtkWidget *widget;
   MetaUIFrame *frame;
   MetaFrameGeometry fgeom;
   XRectangle xrect;
   Region corners_xregion;
   Region window_xregion;
   
-  widget = GTK_WIDGET (frames);
-
   frame = meta_frames_lookup_window (frames, xwindow);
   g_return_if_fail (frame != NULL);
 
@@ -847,7 +934,7 @@ meta_frames_apply_shapes (MetaFrames *frames,
                                     0,
                                     CopyFromParent,
                                     CopyFromParent,
-                                    CopyFromParent,
+                                    (Visual *)CopyFromParent,
                                     CWOverrideRedirect,
                                     &attrs);
 
@@ -917,28 +1004,26 @@ meta_frames_move_resize_frame (MetaFrames *frames,
 			       int         height)
 {
   MetaUIFrame *frame = meta_frames_lookup_window (frames, xwindow);
-  int old_width, old_height;
+  int old_x, old_y, old_width, old_height;
   
   gdk_drawable_get_size (frame->window, &old_width, &old_height);
+  gdk_window_get_position (frame->window, &old_x, &old_y);
 
   gdk_window_move_resize (frame->window, x, y, width, height);
 
   if (old_width != width || old_height != height)
-    gdk_window_invalidate_rect (frame->window, NULL, FALSE);
+    invalidate_whole_window (frames, frame);
 }
 
 void
 meta_frames_queue_draw (MetaFrames *frames,
                         Window      xwindow)
 {
-  GtkWidget *widget;
   MetaUIFrame *frame;
   
-  widget = GTK_WIDGET (frames);
-
   frame = meta_frames_lookup_window (frames, xwindow);
 
-  gdk_window_invalidate_rect (frame->window, NULL, FALSE);
+  invalidate_whole_window (frames, frame);
 }
 
 void
@@ -946,11 +1031,8 @@ meta_frames_set_title (MetaFrames *frames,
                        Window      xwindow,
                        const char *title)
 {
-  GtkWidget *widget;
   MetaUIFrame *frame;
   
-  widget = GTK_WIDGET (frames);
-
   frame = meta_frames_lookup_window (frames, xwindow);
 
   g_assert (frame);
@@ -963,24 +1045,24 @@ meta_frames_set_title (MetaFrames *frames,
       g_object_unref (frame->layout);
       frame->layout = NULL;
     }
-  
-  gdk_window_invalidate_rect (frame->window, NULL, FALSE);
+
+  invalidate_whole_window (frames, frame);
 }
 
 void
 meta_frames_repaint_frame (MetaFrames *frames,
                            Window      xwindow)
 {
-  GtkWidget *widget;
   MetaUIFrame *frame;
   
-  widget = GTK_WIDGET (frames);
-
   frame = meta_frames_lookup_window (frames, xwindow);
 
   g_assert (frame);
 
-  gdk_window_process_updates (frame->window, TRUE);
+  /* repaint everything, so the other frame don't
+   * lag behind if they are exposed
+   */
+  gdk_window_process_all_updates ();
 }
 
 static void
@@ -1119,6 +1201,7 @@ redraw_control (MetaFrames *frames,
   rect = control_rect (control, &fgeom);
 
   gdk_window_invalidate_rect (frame->window, rect, FALSE);
+  invalidate_cache (frames, frame);
 }
 
 static gboolean
@@ -1152,13 +1235,11 @@ meta_frames_button_press_event (GtkWidget      *widget,
   
   control = get_control (frames, frame, event->x, event->y);
 
-  /* raise/focus on click, even if click was on client area */
+  /* focus on click, even if click was on client area */
   if (event->button == 1 &&
       !(control == META_FRAME_CONTROL_MINIMIZE ||
         control == META_FRAME_CONTROL_DELETE))
     {
-      meta_core_user_raise (gdk_display,
-                            frame->xwindow);
       meta_topic (META_DEBUG_FOCUS,
                   "Focusing window with frame 0x%lx due to button 1 press\n",
                   frame->xwindow);
@@ -1200,8 +1281,6 @@ meta_frames_button_press_event (GtkWidget      *widget,
           
         case META_ACTION_DOUBLE_CLICK_TITLEBAR_TOGGLE_MAXIMIZE:
           {
-            MetaFrameFlags flags;
-            
             flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
             
             if (flags & META_FRAME_ALLOWS_MAXIMIZE)
@@ -1257,6 +1336,7 @@ meta_frames_button_press_event (GtkWidget      *widget,
                                frame->xwindow,
                                op,
                                TRUE,
+                               meta_ui_get_last_event_serial (gdk_display),
                                event->button,
                                0,
                                event->time,
@@ -1332,15 +1412,25 @@ meta_frames_button_press_event (GtkWidget      *widget,
           break;
         }
 
-      meta_core_begin_grab_op (gdk_display,
-                               frame->xwindow,
-                               op,
-                               TRUE,
-                               event->button,
-                               0,
-                               event->time,
-                               event->x_root,
-                               event->y_root);
+      if (!meta_core_titlebar_is_onscreen (gdk_display,
+                                           frame->xwindow))
+        meta_core_show_window_menu (gdk_display,
+                                    frame->xwindow,
+                                    event->x_root,
+                                    event->y_root,
+                                    event->button,
+                                    event->time);
+      else
+        meta_core_begin_grab_op (gdk_display,
+                                 frame->xwindow,
+                                 op,
+                                 TRUE,
+                                 meta_ui_get_last_event_serial (gdk_display),
+                                 event->button,
+                                 0,
+                                 event->time,
+                                 event->x_root,
+                                 event->y_root);
     }
   else if (control == META_FRAME_CONTROL_TITLE &&
            event->button == 1)
@@ -1355,6 +1445,7 @@ meta_frames_button_press_event (GtkWidget      *widget,
                                    frame->xwindow,
                                    META_GRAB_OP_MOVING,
                                    TRUE,
+                                   meta_ui_get_last_event_serial (gdk_display),
                                    event->button,
                                    0,
                                    event->time,
@@ -1364,7 +1455,9 @@ meta_frames_button_press_event (GtkWidget      *widget,
     }
   else if (event->button == 2)
     {
-      meta_core_user_lower (gdk_display, frame->xwindow);
+      meta_core_user_lower_and_unfocus (gdk_display,
+                                        frame->xwindow,
+                                        event->time);
     }
   else if (event->button == 3)
     {
@@ -1507,7 +1600,7 @@ meta_frames_update_prelit_control (MetaFrames      *frames,
   MetaFrameControl old_control;
   MetaCursor cursor;
 
-  meta_verbose ("Updating prelit control from %d to %d\n",
+  meta_verbose ("Updating prelit control from %u to %u\n",
                 frame->prelit_control, control);
   
   cursor = META_CURSOR_DEFAULT;
@@ -1681,12 +1774,218 @@ meta_frames_destroy_event           (GtkWidget           *widget,
   return TRUE;
 }
 
+/* Cut and paste from GDK */
+static GdkGC *
+get_bg_gc (GdkWindow *window, int x_offset, int y_offset)
+{
+  GdkWindowObject *private = (GdkWindowObject *)window;
+  guint gc_mask = 0;
+  GdkGCValues gc_values;
+
+  if (private->bg_pixmap == GDK_PARENT_RELATIVE_BG && private->parent)
+    {
+      return get_bg_gc (GDK_WINDOW (private->parent),
+			x_offset + private->x,
+			y_offset + private->y);
+    }
+  else if (private->bg_pixmap && 
+           private->bg_pixmap != GDK_PARENT_RELATIVE_BG && 
+           private->bg_pixmap != GDK_NO_BG)
+    {
+      gc_values.fill = GDK_TILED;
+      gc_values.tile = private->bg_pixmap;
+      gc_values.ts_x_origin = x_offset;
+      gc_values.ts_y_origin = y_offset;
+      
+      gc_mask = GDK_GC_FILL | GDK_GC_TILE | GDK_GC_TS_X_ORIGIN | GDK_GC_TS_Y_ORIGIN;
+
+      return gdk_gc_new_with_values (window, &gc_values, gc_mask);
+    }
+  else
+    {
+      GdkGC *gc = gdk_gc_new (window);
+
+      gdk_gc_set_foreground (gc, &(private->bg_color));
+
+      return gc;
+    }
+}
+
+static void
+clear_backing (GdkPixmap *pixmap,
+	       GdkWindow *window,
+	       int xoffset, int yoffset)
+{
+  GdkGC *tmp_gc = get_bg_gc (window, xoffset, yoffset);
+
+  gdk_draw_rectangle (pixmap, tmp_gc, TRUE,
+		      0, 0, -1, -1);
+  
+  g_object_unref (tmp_gc);
+}
+
+static GdkPixmap *
+generate_pixmap (MetaFrames *frames,
+		 MetaUIFrame *frame,
+		 int x,
+		 int y,
+		 int width,
+		 int height)
+{
+  GdkRectangle rectangle;
+  GdkRegion *region;
+  GdkPixmap *result;
+
+  rectangle.x = x;
+  rectangle.y = y;
+  rectangle.width = MAX (width, 1);
+  rectangle.height = MAX (height, 1);
+  
+  result = gdk_pixmap_new (frame->window, rectangle.width, rectangle.height, -1);
+  
+  clear_backing (result, frame->window, rectangle.x, rectangle.y);
+
+  region = gdk_region_rectangle (&rectangle);
+
+  meta_frames_paint_to_drawable (frames, frame, result, region,
+				 - rectangle.x, - rectangle.y);
+
+  gdk_region_destroy (region);
+
+  return result;
+}
+
+static void
+populate_cache (MetaFrames *frames,
+		MetaUIFrame *frame)
+{
+  int top, bottom, left, right;
+  int width, height;
+  int frame_width, frame_height, screen_width, screen_height;
+  CachedPixels *pixels;
+
+  meta_core_get_frame_extents (gdk_display,
+                               frame->xwindow,
+                               NULL, NULL,
+                               &frame_width, &frame_height);
+
+  meta_core_get_screen_size (gdk_display,
+                             frame->xwindow,
+                             &screen_width, &screen_height);
+
+  /* don't cache extremely large windows */
+  if (frame_width > 2 * screen_width ||
+      frame_height > 2 * screen_height)
+    {
+      return;
+    }
+  
+  meta_theme_get_frame_borders (meta_theme_get_current (),
+                                meta_core_get_frame_type (gdk_display, frame->xwindow),
+				frame->text_height,
+				meta_core_get_frame_flags (gdk_display, frame->xwindow),
+                                &top, &bottom, &left, &right);
+
+  meta_core_get_client_size (gdk_display, frame->xwindow, &width, &height);
+
+  pixels = get_cache (frames, frame);
+  
+  if (!pixels->top)
+    {
+      pixels->top = generate_pixmap (frames, frame,
+				    0, 0, left + width + height, top);
+      pixels->top_x = 0;
+      pixels->top_y = 0;
+    }
+  
+  if (!pixels->left)
+    {
+      pixels->left = generate_pixmap (frames, frame,
+				     0, top, left, height);
+      pixels->left_x = 0;
+      pixels->left_y = top;
+    }
+  
+  if (!pixels->right)
+    {
+      pixels->right = generate_pixmap (frames, frame,
+				      left + width, top, right, height);
+      pixels->right_x = left + width;
+      pixels->right_y = top;
+    }
+  
+  if (!pixels->bottom)
+    {
+      pixels->bottom = generate_pixmap (frames, frame,
+				       0, top + height, left + width + height, bottom);
+      pixels->bottom_x = 0;
+      pixels->bottom_y = top + height;
+    }
+  
+  if (frames->invalidate_cache_timeout_id)
+    g_source_remove (frames->invalidate_cache_timeout_id);
+  
+  frames->invalidate_cache_timeout_id = g_timeout_add (1000, invalidate_cache_timeout, frames);
+
+  if (!g_list_find (frames->invalidate_frames, frame))
+    frames->invalidate_frames =
+      g_list_prepend (frames->invalidate_frames, frame);
+}
+
+static void
+clip_to_screen (GdkRegion *region, MetaUIFrame *frame)
+{
+  GdkRectangle frame_area;
+  GdkRectangle screen_area = { 0, 0, 0, 0 };
+  GdkRegion *tmp_region;
+  
+  /* Chop off stuff outside the screen; this optimization
+   * is crucial to handle huge client windows,
+   * like "xterm -geometry 1000x1000"
+   */
+  meta_core_get_frame_extents (gdk_display,
+                               frame->xwindow,
+                               &frame_area.x, &frame_area.y,
+                               &frame_area.width, &frame_area.height);
+
+  meta_core_get_screen_size (gdk_display,
+                             frame->xwindow,
+                             &screen_area.width, &screen_area.height);
+
+  gdk_region_offset (region, frame_area.x, frame_area.y);
+
+  tmp_region = gdk_region_rectangle (&frame_area);
+  gdk_region_intersect (region, tmp_region);
+  gdk_region_destroy (tmp_region);
+
+  gdk_region_offset (region, - frame_area.x, - frame_area.y);
+}
+
+static void
+subtract_from_region (GdkRegion *region, GdkDrawable *drawable,
+		      gint x, gint y)
+{
+  GdkRectangle rect;
+  GdkRegion *reg_rect;
+
+  gdk_drawable_get_size (drawable, &rect.width, &rect.height);
+  rect.x = x;
+  rect.y = y;
+
+  reg_rect = gdk_region_rectangle (&rect);
+  gdk_region_subtract (region, reg_rect);
+  gdk_region_destroy (reg_rect);
+}
+
 static gboolean
 meta_frames_expose_event (GtkWidget           *widget,
                           GdkEventExpose      *event)
 {
   MetaUIFrame *frame;
   MetaFrames *frames;
+  GdkGC *gc;
+  GdkRegion *region;
+  CachedPixels *pixels;
 
   frames = META_FRAMES (widget);
 
@@ -1701,8 +2000,57 @@ meta_frames_expose_event (GtkWidget           *widget,
       return TRUE;
     }
 
-  meta_frames_paint_to_drawable (frames, frame, frame->window, event->region);
+  populate_cache (frames, frame);
 
+  region = gdk_region_copy (event->region);
+  
+  gc = gdk_gc_new (frame->window);
+
+  pixels = get_cache (frames, frame);
+  
+  if (pixels->top)
+    {
+      gdk_draw_drawable (frame->window, gc, pixels->top,
+			 0, 0,
+			 pixels->top_x, pixels->top_y,
+			 -1, -1);
+      subtract_from_region (region, pixels->top, pixels->top_x, pixels->top_y);
+    }
+
+  if (pixels->left)
+    {
+      gdk_draw_drawable (frame->window, gc, pixels->left,
+			 0, 0,
+			 pixels->left_x, pixels->left_y,
+			 -1, -1);
+      subtract_from_region (region, pixels->left, pixels->left_x, pixels->left_y);
+    }
+
+  if (pixels->right)
+    {
+      gdk_draw_drawable (frame->window, gc, pixels->right,
+			 0, 0,
+			 pixels->right_x, pixels->right_y,
+			 -1, -1);
+      subtract_from_region (region, pixels->right, pixels->right_x, pixels->right_y);
+    }
+  
+  if (pixels->bottom)
+    {
+      gdk_draw_drawable (frame->window, gc, pixels->bottom,
+			 0, 0,
+			 pixels->bottom_x, pixels->bottom_y,
+			 -1, -1);
+      subtract_from_region (region, pixels->bottom, pixels->bottom_x, pixels->bottom_y);
+    }
+
+  clip_to_screen (region, frame);
+  meta_frames_paint_to_drawable (frames, frame,
+				 frame->window, region, 0, 0);
+  
+  g_object_unref (gc);
+  gdk_region_destroy (region);
+  
   return TRUE;
 }
 
@@ -1710,7 +2058,9 @@ static void
 meta_frames_paint_to_drawable (MetaFrames   *frames,
                                MetaUIFrame  *frame,
                                GdkDrawable  *drawable,
-                               GdkRegion    *region)
+                               GdkRegion    *region,
+			       int           x_offset,
+			       int           y_offset)
 {
   GtkWidget *widget;
   MetaFrameFlags flags;
@@ -1727,7 +2077,6 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
   GdkRectangle area;
   GdkRectangle *areas;
   int n_areas;
-  int screen_width, screen_height;
   MetaButtonLayout button_layout;
   MetaGrabOp grab_op;
   
@@ -1824,36 +2173,6 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
   gdk_region_subtract (edges, tmp_region);
   gdk_region_destroy (tmp_region);
 
-  /* Chop off stuff outside the screen; this optimization
-   * is crucial to handle huge client windows,
-   * like "xterm -geometry 1000x1000"
-   */
-  meta_core_get_frame_extents (gdk_display,
-                               frame->xwindow,
-                               &area.x, &area.y,
-                               &area.width, &area.height);
-
-  meta_core_get_screen_size (gdk_display,
-                             frame->xwindow,
-                             &screen_width, &screen_height);
-
-  if ((area.x + area.width) > screen_width)
-    area.width = screen_width - area.x;
-  if (area.width < 0)
-    area.width = 0;
-  
-  if ((area.y + area.height) > screen_height)
-    area.height = screen_height - area.y;
-  if (area.height < 0)
-    area.height = 0;
-
-  area.x = 0; /* make relative to frame rather than screen */
-  area.y = 0;
-  
-  tmp_region = gdk_region_rectangle (&area);
-  gdk_region_intersect (edges, tmp_region);
-  gdk_region_destroy (tmp_region);
-
   /* Now draw remaining portion of region */
   gdk_region_get_rectangles (edges, &areas, &n_areas);
 
@@ -1868,8 +2187,8 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
       meta_theme_draw_frame (meta_theme_get_current (),
                              widget,
                              drawable,
-                             &areas[i],
-                             0, 0,
+                             NULL, /* &areas[i], */
+                             x_offset, y_offset,
                              type,
                              flags,
                              w, h,
@@ -1889,6 +2208,22 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
   g_free (areas);
 }
 
+static void
+meta_frames_set_window_background (MetaFrames   *frames,
+                                   MetaUIFrame  *frame)
+{
+  gtk_style_set_background (GTK_WIDGET (frames)->style,
+                            frame->window, GTK_STATE_NORMAL);
+
+#if 0
+  /* This is what we want for transparent background */
+  {
+    col.pixel = 0;
+    gdk_window_set_background (window, &col);
+  }
+#endif
+}
+
 static gboolean
 meta_frames_enter_notify_event      (GtkWidget           *widget,
                                      GdkEventCrossing    *event)
@@ -2110,6 +2445,13 @@ get_control (MetaFrames *frames,
 void
 meta_frames_push_delay_exposes (MetaFrames *frames)
 {
+  if (frames->expose_delay_count == 0)
+    {
+      /* Make sure we've repainted things */
+      gdk_window_process_all_updates ();
+      XFlush (gdk_display);
+    }
+  
   frames->expose_delay_count += 1;
 }
 
@@ -2124,7 +2466,7 @@ queue_pending_exposes_func (gpointer key, gpointer value, gpointer data)
 
   if (frame->expose_delayed)
     {
-      gdk_window_invalidate_rect (frame->window, NULL, FALSE);
+      invalidate_whole_window (frames, frame);
       frame->expose_delayed = FALSE;
     }
 }
@@ -2143,3 +2485,11 @@ meta_frames_pop_delay_exposes  (MetaFrames *frames)
                             frames);
     }
 }
+
+static void
+invalidate_whole_window (MetaFrames *frames,
+			 MetaUIFrame *frame)
+{
+  gdk_window_invalidate_rect (frame->window, NULL, FALSE);
+  invalidate_cache (frames, frame);
+}
