@@ -4,6 +4,7 @@
  * Copyright (C) 2001 Havoc Pennington
  * Copyright (C) 2002, 2003 Red Hat, Inc.
  * Copyright (C) 2004 Rob Adams
+ * Copyright (C) 2004, 2005 Elijah Newren
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -200,8 +201,7 @@ window_is_fullscreen_size (MetaWindow *window)
 {
   int i;
 
-  if (window->rect.width >= window->screen->width &&
-      window->rect.height >= window->screen->height)
+  if (meta_rectangle_could_fit_rect (&window->rect, &window->screen->rect))
     {
       /* we use the work area since windows that try to
        * position at 0,0 will get pushed down by menu panel
@@ -209,26 +209,20 @@ window_is_fullscreen_size (MetaWindow *window)
       MetaRectangle workarea;
       
       meta_window_get_work_area_current_xinerama (window, &workarea);
-      if (window->rect.x <= workarea.x &&
-          window->rect.y <= workarea.y &&
-	  window->rect.x + window->rect.width >= workarea.x + workarea.width &&
-	  window->rect.y + window->rect.height >= workarea.y + workarea.height) 
+      if (meta_rectangle_contains_rect (&window->rect, &workarea))
         return TRUE;
     }
   
   i = 0;
   while (i < window->screen->n_xinerama_infos)
     {
-      if (window->rect.width >= window->screen->xinerama_infos[i].width &&
-          window->rect.height >= window->screen->xinerama_infos[i].height)
+      if (meta_rectangle_could_fit_rect (&window->rect,
+                                         &window->screen->xinerama_infos[i].rect))
         {
           MetaRectangle workarea;
           
           meta_window_get_work_area_current_xinerama (window, &workarea);
-          if (window->rect.x <= workarea.x &&
-              window->rect.y <= workarea.y &&
-	      window->rect.x + window->rect.width >= workarea.x + workarea.width &&
-	      window->rect.y + window->rect.height >= workarea.y + workarea.height) 
+          if (meta_rectangle_contains_rect (&window->rect, &workarea))
             return TRUE;
         }
       
@@ -242,8 +236,7 @@ static gboolean
 is_focused_foreach (MetaWindow *window,
                     void       *data)
 {
-  if (window->has_focus ||
-      (window == window->display->expected_focus_window))
+  if (window == window->display->expected_focus_window)
     {
       *((gboolean*) data) = TRUE;
       return FALSE;
@@ -251,6 +244,17 @@ is_focused_foreach (MetaWindow *window,
   return TRUE;
 }
 
+static gboolean
+windows_on_different_xinerama (MetaWindow *a,
+                               MetaWindow *b)
+{
+  if (a->screen != b->screen)
+    return TRUE;
+
+  return meta_screen_get_xinerama_for_window (a->screen, a) !=
+    meta_screen_get_xinerama_for_window (b->screen, b);
+}
+
 /* Get layer ignoring any transient or group relationships */
 static MetaStackLayer
 get_standalone_layer (MetaWindow *window)
@@ -272,10 +276,6 @@ get_standalone_layer (MetaWindow *window)
         layer = META_LAYER_DOCK;
       break;
 
-    case META_WINDOW_SPLASHSCREEN:
-      layer = META_LAYER_SPLASH;
-      break;
-      
     default:       
       meta_window_foreach_transient (window,
                                      is_focused_foreach,
@@ -283,9 +283,13 @@ get_standalone_layer (MetaWindow *window)
 
       if (window->wm_state_below)
         layer = META_LAYER_BOTTOM;
-      else if ((window->has_focus || focused_transient ||
-                (window == window->display->expected_focus_window)) &&
-               (window->fullscreen || window_is_fullscreen_size (window)))
+      else if ((window->fullscreen || window_is_fullscreen_size (window)) &&
+               (focused_transient ||
+                window == window->display->expected_focus_window ||
+                window->display->expected_focus_window == NULL ||
+                (window->display->expected_focus_window != NULL &&
+                 windows_on_different_xinerama (window,
+                                                window->display->expected_focus_window))))
         layer = META_LAYER_FULLSCREEN;
       else if (window->wm_state_above)
         layer = META_LAYER_DOCK;
@@ -363,13 +367,13 @@ compute_layer (MetaWindow *window)
       if (group_max > window->layer)
         {
           meta_topic (META_DEBUG_STACK,
-                      "Promoting window %s from layer %d to %d due to group membership\n",
+                      "Promoting window %s from layer %u to %u due to group membership\n",
                       window->desc, window->layer, group_max);
           window->layer = group_max;
         }
     }
 
-  meta_topic (META_DEBUG_STACK, "Window %s on layer %d type = %d has_focus = %d\n",
+  meta_topic (META_DEBUG_STACK, "Window %s on layer %u type = %u has_focus = %d\n",
               window->desc, window->layer,
               window->type, window->has_focus);
 }
@@ -663,7 +667,7 @@ ensure_above (MetaWindow *above,
       above->layer < below->layer)
     {
       meta_topic (META_DEBUG_STACK,
-		  "Promoting window %s from layer %d to %d due to contraint\n",
+		  "Promoting window %s from layer %u to %u due to contraint\n",
 		  above->desc, above->layer, below->layer);
       above->layer = below->layer;
     }
@@ -883,7 +887,7 @@ meta_stack_ensure_sorted (MetaStack *stack)
           if (w->layer != old_layer)
             {
               meta_topic (META_DEBUG_STACK,
-                          "Window %s moved from layer %d to %d\n",
+                          "Window %s moved from layer %u to %u\n",
                           w->desc, old_layer, w->layer);
               
               stack->need_resort = TRUE;
@@ -935,7 +939,7 @@ raise_window_relative_to_managed_windows (MetaScreen *screen,
 
   Window ignored1, ignored2;
   Window *children;
-  int n_children;
+  unsigned int n_children;
   int i;
 
   /* Normally XQueryTree() means "must grab server" but here
@@ -1059,7 +1063,7 @@ meta_stack_sync_to_server (MetaStack *stack)
       else
         g_array_append_val (root_children_stacked, w->xwindow);
       
-      meta_topic (META_DEBUG_STACK, "%d:%d - %s ", w->layer, w->stack_position, w->desc);
+      meta_topic (META_DEBUG_STACK, "%u:%d - %s ", w->layer, w->stack_position, w->desc);
           
       tmp = tmp->next;
     }
@@ -1069,12 +1073,12 @@ meta_stack_sync_to_server (MetaStack *stack)
 
   /* All windows should be in some stacking order */
   if (stacked->len != stack->windows->len)
-    meta_bug ("%d windows stacked, %d windows exist in stack\n",
+    meta_bug ("%u windows stacked, %u windows exist in stack\n",
               stacked->len, stack->windows->len);
   
   /* Sync to server */
 
-  meta_topic (META_DEBUG_STACK, "Restacking %d windows\n",
+  meta_topic (META_DEBUG_STACK, "Restacking %u windows\n",
               root_children_stacked->len);
   
   meta_error_trap_push (stack->screen->display);
@@ -1344,7 +1348,7 @@ get_default_focus_window (MetaStack     *stack,
           !window->minimized &&
           (window->input || window->take_focus) &&
           (workspace == NULL ||
-           meta_window_visible_on_workspace (window, workspace)))
+           meta_window_located_on_workspace (window, workspace)))
         {
           if (topmost_dock == NULL &&
               window->type == META_WINDOW_DOCK)
@@ -1431,7 +1435,7 @@ meta_stack_list_windows (MetaStack     *stack,
       MetaWindow *window = link->data;
       
       if (window &&
-          (workspace == NULL || meta_window_visible_on_workspace (window, workspace)))
+          (workspace == NULL || meta_window_located_on_workspace (window, workspace)))
         {
           workspace_windows = g_list_prepend (workspace_windows,
                                               window);
@@ -1466,6 +1470,110 @@ meta_stack_windows_cmp  (MetaStack  *stack,
     return 0; /* not reached */
 }
 
+static int
+compare_just_window_stack_position (void *a,
+                                    void *b)
+{
+  MetaWindow *window_a = a;
+  MetaWindow *window_b = b;
+
+  if (window_a->stack_position < window_b->stack_position)
+    return -1; /* move window_a earlier in list */
+  else if (window_a->stack_position > window_b->stack_position)
+    return 1;
+  else
+    return 0; /* not reached */
+}
+
+GList*
+meta_stack_get_positions (MetaStack *stack)
+{
+  GList *tmp;
+
+  /* Make sure to handle any adds or removes */
+  meta_stack_ensure_sorted (stack);
+
+  tmp = g_list_copy (stack->sorted);
+  tmp = g_list_sort (tmp, (GCompareFunc) compare_just_window_stack_position);
+
+  return tmp;
+}
+
+static gint
+compare_pointers (gconstpointer a,
+                  gconstpointer b)
+{
+  if (a > b)
+    return 1;
+  else if (a < b)
+    return -1;
+  else 
+    return 0;
+}
+
+static gboolean
+lists_contain_same_windows (GList *a,
+                            GList *b)
+{
+  GList *copy1, *copy2;
+  GList *tmp1, *tmp2;
+
+  if (g_list_length (a) != g_list_length (b))
+    return FALSE;
+
+  tmp1 = copy1 = g_list_sort (g_list_copy (a), compare_pointers);
+  tmp2 = copy2 = g_list_sort (g_list_copy (b), compare_pointers);
+
+  while (tmp1 && tmp1->data == tmp2->data)   /* tmp2 is non-NULL if tmp1 is */
+    {
+      tmp1 = tmp1->next;
+      tmp2 = tmp2->next;
+    }
+
+  g_list_free (copy1);
+  g_list_free (copy2);
+
+  return (tmp1 == NULL);    /* tmp2 is non-NULL if tmp1 is */
+}
+
+void
+meta_stack_set_positions (MetaStack *stack,
+                          GList     *windows)
+{
+  int i;
+  GList *tmp;
+
+  /* Make sure any adds or removes aren't in limbo -- is this needed? */
+  meta_stack_ensure_sorted (stack);
+  
+  if (!lists_contain_same_windows (windows, stack->sorted))
+    {
+      meta_warning ("This list of windows has somehow changed; not resetting "
+                    "positions of the windows.\n");
+      return;
+    }
+
+  g_list_free (stack->sorted);
+  stack->sorted = g_list_copy (windows);
+
+  stack->need_resort = TRUE;
+  stack->need_constrain = TRUE;
+   
+  i = 0;
+  tmp = windows;
+  while (tmp != NULL)
+    {
+      MetaWindow *w = tmp->data;
+      w->stack_position = i++;
+      tmp = tmp->next;
+    }
+  
+  meta_topic (META_DEBUG_STACK,
+              "Reset the stack positions of (nearly) all windows\n");
+
+  meta_stack_sync_to_server (stack);
+}
+
 void
 meta_window_set_stack_position_no_sync (MetaWindow *window,
                                         int         position)
